/* request_firmware() hotplug interface
   (C) Copyright Manuel Estrada Sainz 2003
   
   Based on kernel/kmod.c and drivers/usb/usb.c
*/
/*
        kernel/kmod.c
        Kirk Petersen

        Reorganized not to be a daemon by Adam Richter, with guidance
        from Greg Zornetzer.

        Modified to avoid chroot and file sharing problems.
        Mikael Pettersson

        Limit the concurrent number of kmod modprobes to catch loops from
        "modprobe needs a service that is in a module".
        Keith Owens <kaos@ocs.com.au> December 1999

        Unblock all signals when we exec a usermode process.
        Shuu Yamaguchi <shuu@wondernetworkresources.com> December 2000
*/
/*
 * drivers/usb/usb.c
 *
 * (C) Copyright Linus Torvalds 1999
 * (C) Copyright Johannes Erdfelt 1999-2001
 * (C) Copyright Andreas Gal 1999
 * (C) Copyright Gregory P. Smith 1999
 * (C) Copyright Deti Fliegl 1999 (new USB architecture)
 * (C) Copyright Randy Dunlap 2000
 * (C) Copyright David Brownell 2000 (kernel hotplug, usb_device_id)
 * (C) Copyright Yggdrasil Computing, Inc. 2000
 *     (usb_device_id matching changes by Adam J. Richter)
 */

#define __KERNEL_SYSCALLS__

#include <linux/config.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/pagemap.h>
#include <linux/init.h>
#include <linux/string.h>
#include <linux/types.h>
#include <linux/slab.h>
#include <linux/spinlock.h>
#include <asm/uaccess.h>
#include <linux/kmod.h>
#include <linux/unistd.h>


#include "firmware.h"
#include "fwfs.h"
#define DEBUG  
#include <linux/usb.h>

MODULE_LICENSE("GPL");

struct firmware_exec {
	char **argv;
	char **envp;
};
static int errno;
static int exec_hotplug (void *arg)
{
	struct firmware_exec *fw_exec = arg;
	int ret;

        ret = exec_usermodehelper (fw_exec->argv[0],
				   fw_exec->argv, fw_exec->envp);
	if (ret) {
                printk(KERN_ERR "firmware: failed to exec %s, ret = %d\n",
                       hotplug_path, ret);
        }
		
	return ret;
}
static int call_helper (char *verb, const char *name, const char *device)
{
	char *argv [3], **envp, *buf, *scratch;
        int i = 0;

        pid_t pid;
        int waitpid_result;
	int retval=0;
        sigset_t tmpsig;
	struct firmware_exec fw_exec;
	
        if (!hotplug_path [0])
                return -ENOENT;
        if (in_interrupt ()) {
                dbg ("In_interrupt");
                return -EFAULT;
        }
        if (!current->fs->root) {
                /* statically linked USB is initted rather early */
                dbg ("call_policy %s -- no FS yet", verb);
                return -EPERM;
        }

        if (!(envp = (char **) kmalloc (20 * sizeof (char *), GFP_KERNEL))) {
                dbg ("enomem");
                return -ENOMEM;
        }
        if (!(buf = kmalloc (256, GFP_KERNEL))) {
                kfree (envp);
                dbg ("enomem2");
                return -ENOMEM;
        }

        /* only one standardized param to hotplug command: type */
        argv [0] = hotplug_path;
        argv [1] = "firmware";
        argv [2] = 0;

        /* minimal command environment */
        envp [i++] = "HOME=/";
        envp [i++] = "PATH=/sbin:/bin:/usr/sbin:/usr/bin";

#ifdef  DEBUG
        /* hint that policy agent should enter no-stdout debug mode */
        envp [i++] = "DEBUG=kernel";
#endif
        /* extensible set of named bus-specific parameters,
         * supporting multiple driver selection algorithms.
         */
        scratch = buf;

	if(device){
		envp [i++] = scratch;
		scratch += sprintf (scratch, "DEVICE=%s", device) + 1;
	}
        /* action:  load, unload */
        envp [i++] = scratch;
        scratch += sprintf (scratch, "ACTION=%s", verb) + 1;

        envp [i++] = scratch;
        scratch += sprintf (scratch, "FIRMWARE=%s", name) + 1;

        envp [i++] = 0;
        /* assert: (scratch - buf) < sizeof buf */

        /* NOTE: user mode daemons can call the agents too */

        dbg ("firmware: %s %s", argv [0], verb);

	fw_exec.argv = argv;
	fw_exec.envp = envp;

        pid = kernel_thread(exec_hotplug, (void *)&fw_exec, 0);
        if (pid < 0) {
                printk(KERN_ERR "request_firmware[%s]:"
		       " fork failed, errno %d\n", name, -pid);
		retval=pid;
                goto out;
        }

        /* Block everything but SIGKILL/SIGSTOP */
        spin_lock_irq(&current->sigmask_lock);
        tmpsig = current->blocked;
        siginitsetinv(&current->blocked, sigmask(SIGKILL) | sigmask(SIGSTOP));
        recalc_sigpending(current);
        spin_unlock_irq(&current->sigmask_lock);

        waitpid_result = waitpid(pid, NULL, __WCLONE);

        /* Allow signals again.. */
        spin_lock_irq(&current->sigmask_lock);
        current->blocked = tmpsig;
        recalc_sigpending(current);
        spin_unlock_irq(&current->sigmask_lock);

        if (waitpid_result != pid) {
                printk(KERN_ERR "request_firmware[%s]:"
		       " waitpid(%d,...) failed, errno %d\n",
                       name, pid, -waitpid_result);
		retval = -waitpid_result;
        }
out:
        kfree (buf);
        kfree (envp);
	return retval;
}
int request_firmware (const struct firmware **firmware, const char *name,
		      const char *device)
/* For 2.5 'device' should be 'struct device *' */
{
	struct firmware *fw;
	int retval;
	
	retval = call_helper("load", name, device);

	if (firmware){
		*firmware=NULL;
		fw = fwfs_get(name);
		if(!fw)
			return -ENOENT;
		strncpy(fw->name, name, FIRMWARE_NAME_MAX);
		fw->name[FIRMWARE_NAME_MAX-1] = '\0';
		*firmware=fw;
	}
	return retval;
}
void release_firmware (const struct firmware *fw)
{
	call_helper("unload", fw->name, NULL);
	fwfs_put(fw);
}
void register_firmware (const char *name, const u8 *data, size_t size)
{
	fwfs_write_default(name, data, size);
	call_helper("register", name, NULL);
}
